# -*- coding: utf-8 -*-
"""
alembic handlers params module.
"""

import pyrin.globalization.datetime.services as datetime_services
import pyrin.configuration.services as config_services

from pyrin.cli.arguments import BooleanArgument, KeywordArgument, PositionalArgument
from pyrin.database.migration.alembic.interface import AlembicCLIParamBase


class SQLParam(BooleanArgument, AlembicCLIParamBase):
    """
    sql param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SQLParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('sql', '--sql', default=default)


class TagParam(KeywordArgument, AlembicCLIParamBase):
    """
    tag param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of TagParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('tag', '--tag', default=default)


class RevisionParam(PositionalArgument, AlembicCLIParamBase):
    """
    revision param class.
    """

    def __init__(self, index=None, default=None, **options):
        """
        initializes an instance of RevisionParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.

        :keyword bool validate_index: specifies that index of this argument
                                      must be validated. it could be helpful
                                      to set this to False when there are multiple
                                      arguments with the same index that will appear
                                      in different situations.
                                      defaults to True if not provided.
        """

        if index is None:
            index = 0

        super().__init__('revision', index, default=default, **options)


class ResolveDependenciesParam(BooleanArgument, AlembicCLIParamBase):
    """
    resolve dependencies param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of ResolveDependenciesParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('resolve_dependencies', '--resolve-dependencies', default=default)


class IndicateCurrentParam(BooleanArgument, AlembicCLIParamBase):
    """
    indicate current param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of IndicateCurrentParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('indicate_current', '--indicate-current', default=default)


class RevisionRangeParam(KeywordArgument, AlembicCLIParamBase):
    """
    revision range param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of RevisionRangeParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('revision_range', '--rev-range', default=default)


class MessageParam(KeywordArgument, AlembicCLIParamBase):
    """
    message param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MessageParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('message', '--message', default=default)

    def _process_inputs(self, **options):
        """
        processes the inputs in given dict.

        :rtype: dict
        """

        if options.get('message', None) is None:
            timezone = config_services.get('alembic', 'alembic', 'timezone')
            message = datetime_services.get_current_timestamp(date_sep=None,
                                                              main_sep=None,
                                                              time_sep=None,
                                                              timezone=timezone)
            options.update(message=message)

        return super()._process_inputs(**options)


class BranchLabelParam(KeywordArgument, AlembicCLIParamBase):
    """
    branch label param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of BranchLabelParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('branch_label', '--branch-label', default=default)


class RevisionIDParam(KeywordArgument, AlembicCLIParamBase):
    """
    revision id param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of RevisionIDParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('revision_id', '--rev-id', default=default)


class RevisionsParam(PositionalArgument, AlembicCLIParamBase):
    """
    revisions param class.
    """

    def __init__(self, index=None, default=None, **options):
        """
        initializes an instance of RevisionsParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.

        :keyword bool validate_index: specifies that index of this argument
                                      must be validated. it could be helpful
                                      to set this to False when there are multiple
                                      arguments with the same index that will appear
                                      in different situations.
                                      defaults to True if not provided.
        """

        if index is None:
            index = 0

        super().__init__('revisions', index, default=default, **options)


class AutoGenerateParam(BooleanArgument, AlembicCLIParamBase):
    """
    autogenerate param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AutoGenerateParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('autogenerate', '--autogenerate', default=default)


class HeadParam(KeywordArgument, AlembicCLIParamBase):
    """
    head param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of HeadParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('head', '--head', default=default)


class SpliceParam(BooleanArgument, AlembicCLIParamBase):
    """
    splice param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SpliceParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('splice', '--splice', default=default)


class VersionPathParam(KeywordArgument, AlembicCLIParamBase):
    """
    version path param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of VersionPathParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('version_path', '--version-path', default=default)


class DependsOnParam(KeywordArgument, AlembicCLIParamBase):
    """
    depends on param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DependsOnParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('depends_on', '--depends-on', default=default)


class PurgeParam(BooleanArgument, AlembicCLIParamBase):
    """
    purge param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of PurgeParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to None if not provided.
        """

        super().__init__('purge', '--purge', default=default)



# -*- coding: utf-8 -*-
"""
sequence manager module.
"""

from abc import abstractmethod

from sqlalchemy import Sequence

import pyrin.database.services as database_services

from pyrin.core.structs import Manager, DTO
from pyrin.core.exceptions import CoreNotImplementedError
from pyrin.database.sequence import SequencePackage
from pyrin.database.services import get_current_store


class SequenceManager(Manager):
    """
    sequence manager class.
    """

    package_class = SequencePackage

    def __init__(self):
        """
        initializes an instance of SequenceManager.
        """

        super().__init__()

        # a dictionary containing dialect name to method map.
        # different dialects may not all support the standard sequence
        # generation, so we need to provide different methods on different
        # dialects. for example on `mssql` and `sqlite`
        # in the form of: {str dialect_name: callable method}
        self._dialect_to_method_map = self._map_dialect_to_method()

    def get_next_value(self, entity_class, sequence):
        """
        gets the next value of given sequence.

        :param BaseEntity entity_class: entity class which this
                                        sequence is defined in.

        :param Sequence sequence: sequence object to get its next value.

        :rtype: int
        """

        engine = database_services.get_entity_engine(entity_class)

        method = self._dialect_to_method_map.get(engine.name, )

    def _default_next_value(self, engine, sequence):
        """
        gets the next value of given sequence.

        :param Sequence sequence: sequence object to get its next value.

        :rtype: int
        """

        store = get_current_store()
        seq = Sequence(name)

        return store.execute(seq)

    def _mssql_next_value(self, sequence):
        """
        gets the next value of given sequence.
        this method will be used on sql-server backend.

        :param Sequence sequence: sequence object to get its next value.

        :rtype: int
        """

        store = get_current_store()

        return store.execute('select next value for {sequence}'.
                             format(sequence=name)).scalar()

    @abstractmethod
    def _sqlite_next_value(self, sequence):
        """
        sqlite does not support sequences.
        so this method raises an error.

        :param Sequence sequence: sequence object to get its next value.

        :raises CoreNotImplementedError: core not implemented error.
        """

        raise CoreNotImplementedError('sqlite database does not support sequences.')

    def _map_dialect_to_method(self):
        """
        maps different dialects to relevant methods.

        :returns: dict[str dialect_name: callable method]
        :rtype: dict
        """

        result = DTO(default=self._default_next_value,
                     mssql=self._mssql_next_value,
                     sqlite=self._sqlite_next_value)

        return result



# -*- coding: utf-8 -*-
"""
celery structs module.
"""

from celery import Task

import pyrin.logging.services as logging_services


class ExtendedTask(Task):
    """
    extended task class.

    this class adds support for suppressing errors on broker connection failure.
    """

    LOGGER = logging_services.get_logger('celery')

    def apply_async(self, args=None, kwargs=None, task_id=None, producer=None,
                    link=None, link_error=None, shadow=None, **options):
        """
        executes `apply_async` method on task.

        it could suppress if any error occurred on connecting to broker.
        all extra keyword arguments of original `apply_async` method will be passed
        to underlying method.

        :param tuple args: the positional arguments to pass on to the task.
        :param dict kwargs: the keyword arguments to pass on to the task.
        :param str task_id: task id.
        :param kombu.Producer producer: custom producer to use when publishing the task.

        :param inspect.Signature link: a single, or a list of tasks signatures
                                       to apply if the task returns successfully.

        :param inspect.Signature link_error: a single, or a list of task signatures
                                             to apply if an error occurs while executing
                                             the task.

        :param str shadow: override task name used in logs/monitoring.

        :keyword bool suppress: suppress any error during execution of the task
                                and silently logs it. defaults to False if not provided.

        :returns: the `apply_async` result
        """

        suppress = options.pop('suppress', False)

        try:
            return super().apply_async(args=args, kwargs=kwargs, task_id=task_id,
                                       producer=producer, link=link, link_error=link_error,
                                       shadow=shadow, **options)

        except Exception as error:
            if suppress is True:
                self.LOGGER.exception(str(error))
                return None

            raise

    def delay(self, *args, **kwargs):
        """
        executes `delay` method on task.

        it could suppress if any error occurred on connecting to broker.
        all extra keyword arguments of original `delay` method will be passed to
        underlying method.

        :param object args: the positional arguments to pass on to the task.
        :param object kwargs: the keyword arguments to pass on to the task.

        :keyword bool suppress: suppress any error during execution of the task
                                and silently logs it. defaults to False if not provided.

        :returns: the `delay` result
        """

        suppress = kwargs.pop('suppress', False)
        return self.apply_async(args=args, kwargs=kwargs, suppress=suppress)



# -*- coding: utf-8 -*-
"""
validator handlers datetime module.
"""

from datetime import datetime, date, time

import pyrin.globalization.datetime.services as datetime_services

from pyrin.core.globals import _
from pyrin.validator.handlers.base import ValidatorBase
from pyrin.validator.handlers.exceptions import ValueIsNotDateTimeError, ValueIsNotDateError, \
    ValueIsNotTimeError


class DateTimeValidator(ValidatorBase):
    """
    datetime validator class.
    """

    invalid_type_error = ValueIsNotDateTimeError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a datetime.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of DateTimeValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=datetime)
        super().__init__(domain, name, **options)

    def _get_safe_representation(self, value):
        """
        gets the string representable version of input value.

        :param datetime value: value to get its string representation.

        :returns: string representable value.
        :rtype: str
        """

        return datetime_services.to_datetime_string(value, server=False)


class DateValidator(ValidatorBase):
    """
    date validator class.
    """

    invalid_type_error = ValueIsNotDateError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a date.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of DateValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=date)
        super().__init__(domain, name, **options)

    def _get_safe_representation(self, value):
        """
        gets the string representable version of input value.

        :param date value: value to get its string representation.

        :returns: string representable value.
        :rtype: str
        """

        return datetime_services.to_date_string(value)


class TimeValidator(ValidatorBase):
    """
    time validator class.
    """

    invalid_type_error = ValueIsNotTimeError
    invalid_type_message = _('The provided value for [{param_name}] '
                             'must be a time.')

    def __init__(self, domain, name, **options):
        """
        initializes an instance of TimeValidator.

        :param type[BaseEntity] | str domain: the domain in which this validator
                                              must be registered. it could be a
                                              type of a BaseEntity subclass.
                                              if a validator must be registered
                                              independent from any BaseEntity subclass,
                                              the domain could be a unique string name.
                                              note that the provided string name must be
                                              unique at application level.

        :param str name: validator name.
                         each validator will be registered with its name
                         in corresponding domain.
                         to enable automatic validations, the provided
                         name must be the exact name of the parameter
                         which this validator will validate.

        :keyword bool nullable: specifies that null values should be accepted as valid.
                                defaults to True if not provided.

        :keyword str localized_name: localized name of the parameter
                                     which this validator will validate.
                                     it must be passed using `_` method
                                     from `pyrin.core.globals`.
                                     defaults to `name` if not provided.

        :keyword bool is_list: specifies that the value must be a list of items.
                               defaults to False if not provided.

        :keyword bool null_items: specifies that list items could be None.
                                  it is only used if `is_list=True` is provided.
                                  defaults to False if not provided.

        :keyword bool allow_single: specifies that list validator should also
                                    accept single, non list values.
                                    it is only used if `is_list=True` is provided.
                                    defaults to False if not provided.

        :keyword bool allow_empty_list: specifies that list validators should also
                                        accept empty lists.
                                        it is only used if `is_list=True` is provided.
                                        defaults to False if not provided.

        :raises ValidatorNameIsRequiredError: validator name is required error.
        :raises InvalidValidatorDomainError: invalid validator domain error.
        :raises ValidatorFixerMustBeCallable: validator fixer must be callable.
        :raises InvalidValidationExceptionTypeError: invalid validation exception type error.
        """

        options.update(accepted_type=time)
        super().__init__(domain, name, **options)

    def _get_safe_representation(self, value):
        """
        gets the string representable version of input value.

        :param time value: value to get its string representation.

        :returns: string representable value.
        :rtype: str
        """

        return datetime_services.to_time_string(value, server=False)



# -*- coding: utf-8 -*-
"""
celery cli handlers inspect module.
"""

from pyrin.task_queues.celery.cli.decorators import celery_cli_handler
from pyrin.task_queues.celery.cli.enumerations import CeleryCLIHandlersEnum
from pyrin.task_queues.celery.cli.interface import CeleryCLIHandlerBase
from pyrin.task_queues.celery.cli.handlers.params import InspectMethodParam, \
    TimeoutParam, DestinationParam, JSONParam, IncludeDefaultsParam, SamplesCountParam, \
    ObjectTypeParam, CountParam, MaxDepthParam, TaskIDListParam, AttributeListParam


@celery_cli_handler()
class InspectCLIHandler(CeleryCLIHandlerBase):
    """
    inspect cli handler class.
    """

    def __init__(self):
        """
        initializes an instance of InspectCLIHandler.
        """

        super().__init__(CeleryCLIHandlersEnum.INSPECT)

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([InspectMethodParam(index=0), IncludeDefaultsParam(index=1),
                       SamplesCountParam(index=1, validate_index=False),
                       ObjectTypeParam(index=1, validate_index=False),
                       TaskIDListParam(index=1, validate_index=False),
                       AttributeListParam(index=1, validate_index=False),
                       CountParam(index=2), MaxDepthParam(index=3),
                       TimeoutParam(), DestinationParam(), JSONParam()])

        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
logging decorators module.
"""

import time

from functools import update_wrapper

import pyrin.utils.function as func_utils
import pyrin.logging.services as logging_services
import pyrin.configuration.services as config_services


def audit(func):
    """
    decorator to log execution time of a function.

    :param function func: function.

    :returns: function result.
    """

    def decorator(*args, **kwargs):
        """
        decorates the given function and logs its execution time.

        note that `audit_log: true` is required in logging config
        store for each environment to enable this decorator.

        :param object args: function arguments.
        :param object kwargs: function keyword arguments.

        :returns: function result.
        """

        if config_services.get_active('logging', 'audit_log') is not True:
            return func(*args, **kwargs)

        start_time = time.time()
        try:
            return func(*args, **kwargs)

        finally:
            end_time = time.time()
            logging_services.debug('Duration of function call [{name}]: [{time} ms].'
                                   .format(name=func_utils.get_fully_qualified_name(func),
                                           time='{:0.5f}'
                                           .format((end_time - start_time) * 1000)))

    return update_wrapper(decorator, func)


def logging_hook():
    """
    decorator to register a logging hook.

    :raises InvalidLoggingHookTypeError: invalid logging hook type error.

    :returns: logging hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available logging hooks.

        :param type cls: logging hook class.

        :returns: logging hook class.
        :rtype: type
        """

        instance = cls()
        logging_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
logging hooks module.
"""

import pyrin.logging.services as logging_services

from pyrin.core.structs import Hook
from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


class LoggingHookBase(Hook):
    """
    logging hook base class.

    all packages that need to be hooked into logging business must
    implement this class and register it in logging hooks.
    """

    def prepare_data(self, data, **options):
        """
        this method will be called when a log message interpolation is required.

        each subclass must return the modified or same input data.

        :param dict | object data: data that is passed to logging method.

        :returns: modified or same input data.
        :rtype: dict | object
        """

        return data

    def before_emit(self, message, data, level, **options):
        """
        this method will be called before a log is emitted.

        :param str message: the log message that must be emitted.
        :param dict | object data: data that is passed to logging method.
        :param int level: log level.
        """
        pass

    def after_emit(self, message, data, level, **options):
        """
        this method will be called after a log is emitted.

        :param str message: the log message that has been emitted.
        :param dict | object data: data that is passed to logging method.
        :param int level: log level.
        """
        pass


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def after_packages_loaded(self):
        """
        this method will be called after all application packages have been loaded.
        """

        # we must wrap all available loggers into an adapter
        # to inject request info into every log record.
        # it does not affect sqlalchemy logs.
        logging_services.wrap_all_loggers()



# -*- coding: utf-8 -*-
"""
string normalizer handlers space module.
"""

from pyrin.utilities.string.normalizer.decorators import string_normalizer
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.handlers.base import StringNormalizerBase


@string_normalizer()
class SpaceNormalizer(StringNormalizerBase):
    """
    space normalizer class.

    this normalizer removes all spaces.
    """

    def __init__(self, **options):
        """
        initializes an instance of SpaceNormalizer.
        """

        super().__init__(NormalizerEnum.SPACE, 120, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        return value.replace(' ', '')


@string_normalizer()
class DuplicateSpaceNormalizer(StringNormalizerBase):
    """
    duplicate space normalizer class.

    this normalizer replaces all duplicate spaces with a single one.
    """

    def __init__(self, **options):
        """
        initializes an instance of DuplicateSpaceNormalizer.
        """

        super().__init__(NormalizerEnum.DUPLICATE_SPACE, 110, **options)

    def _normalize(self, value, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :returns: normalized value.
        :rtype: str
        """

        while '  ' in value:
            value = value.replace('  ', ' ')

        return value



# -*- coding: utf-8 -*-
"""
database paging component module.
"""

from pyrin.application.decorators import component
from pyrin.application.structs import Component
from pyrin.database.paging import DatabasePagingPackage
from pyrin.database.paging.manager import DatabasePagingManager


@component(DatabasePagingPackage.COMPONENT_NAME)
class DatabasePagingComponent(Component, DatabasePagingManager):
    """
    database paging component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
packaging component module.
"""

from pyrin.application.structs import Component
from pyrin.packaging.manager import PackagingManager


class PackagingComponent(Component, PackagingManager):
    """
    packaging component class.

    an instance of this class will be registered into application components
    on startup, and we shouldn't use `@component` decorator for this class.
    """
    pass



# -*- coding: utf-8 -*-
"""
cli package.
this package provides facilities to execute commands in shell.
"""

from pyrin.packaging.base import Package


class CLIPackage(Package):
    """
    cli package class.
    """

    NAME = __name__
    DEPENDS = []
    COMPONENT_NAME = 'cli.component'



# -*- coding: utf-8 -*-
"""
orm sql extractor services module.
"""

from pyrin.application.services import get_component
from pyrin.database.orm.sql.extractor import ORMSQLExtractorPackage


def find_table_names(expression, **options):
    """
    finds table names from a string or `TextClause` sql expression.

    :param str | TextClause expression: a string or `TextClause`
                                        containing a sql expression.

    :keyword bool include_select: specifies that select statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_insert: specifies that insert statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_update: specifies that update statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :keyword bool include_delete: specifies that delete statements
                                  must be investigated for table names.
                                  defaults to True if not provided.

    :rtype: list[str]
    """

    return get_component(ORMSQLExtractorPackage.COMPONENT_NAME).find_table_names(expression,
                                                                                 **options)



# -*- coding: utf-8 -*-
"""
model services module.
"""

from pyrin.application.services import get_component
from pyrin.database.model import ModelPackage


def get_declarative_base():
    """
    gets the application's declarative base class.

    by default it returns `CoreEntity` class.
    if you want to define a new declarative base class in your application
    instead of using `CoreEntity` which is provided by pyrin, you must
    override this method inside your application's `database.model.ModelManager`
    class and return your new declarative base class.
    this is required if you want to enable migrations in your application.
    note that your application must have a unique declarative base type and
    you shouldn't mix the usage of `CoreEntity` and your new declarative base,
    otherwise migrations will not work properly.

    but, if you don't want to use migrations at all, you could just put
    `pyrin.database.migration` into `ignored_packages` list of `packaging.ini`
    file and leave this method unimplemented.

    :rtype: type[CoreEntity]
    """

    return get_component(ModelPackage.COMPONENT_NAME).get_declarative_base()



# -*- coding: utf-8 -*-
"""
packaging base module.
"""

from pyrin.core.structs import CoreObject
from pyrin.settings.static import DEFAULT_COMPONENT_KEY


class Package(CoreObject):
    """
    package base class.

    all application python packages should be subclassed from this.
    except some base packages like `application`, `core` and `utils` that
    should not implement Package class.
    """

    # the name of the package.
    # example: `pyrin.api`.
    # should get it from `__name__` for each package.
    NAME = None

    # list of all packages that this package is dependent
    # on them and should be loaded after all of them.
    # example: ['pyrin.logging', 'pyrin.configuration']
    # notice that all dependencies on `pyrin.application`
    # and `pyrin.packaging` should not be added into this list
    # because those two packages will be loaded at the beginning
    # and are always available before any other package gets loaded.
    DEPENDS = []

    # specifies that this package is enabled and must be loaded.
    ENABLED = True

    # component name should be unique for each package unless it's intended
    # to replace an already available one. it should be fully qualified name
    # omitting the root application package name (for example `pyrin`), pointing to
    # a real file inside the package (omitting `.py` extension).
    # for example: `api.component` or `security.session.component` or 'my_package.my_component'
    # packages that need to extend (replace) an already available package, must
    # be subclassed from parent package and should not set this attribute.
    COMPONENT_NAME = None

    # component custom key should be unique for each instance unless it's intended
    # to replace an already available one.
    # custom key usage is when we want to expose different implementations
    # based on request context.
    COMPONENT_CUSTOM_KEY = DEFAULT_COMPONENT_KEY

    # all configuration stores that should be loaded automatically by this package.
    # note that the relevant config file for them will also be created in application
    # settings path based on pyrin default setting files if not available.
    CONFIG_STORE_NAMES = []

    # all configuration stores that should not be loaded automatically by this package.
    # note that the relevant config file for them will be created in application
    # settings path based on pyrin default setting files if not available.
    EXTRA_CONFIG_STORE_NAMES = []

    # notice that any package that adds values in either of `CONFIG_STORE_NAMES` or
    # `EXTRA_CONFIG_STORE_NAMES` must also add `pyrin.configuration` into its `DEPENDS` list.

    def load_configs(self, config_services):
        """
        loads all required configs of this package.

        :param Module config_services: configuration services dependency.
                                       to be able to overcome circular dependency problem,
                                       we should inject configuration services dependency
                                       into this method. because all other packages are
                                       referenced `packaging.base` module in them, so we
                                       can't import `pyrin.configuration.services` in this
                                       module. this is more beautiful in comparison to
                                       importing it inside this method.
        """

        if len(self.CONFIG_STORE_NAMES) > 0:
            config_services.load_configurations(*self.CONFIG_STORE_NAMES,
                                                defaults=self.config_defaults,
                                                ignore_on_existed=True)

        if len(self.EXTRA_CONFIG_STORE_NAMES) > 0:
            config_services.create_config_files(*self.EXTRA_CONFIG_STORE_NAMES,
                                                ignore_on_existed=True,
                                                silent=True)

        self._load_configs(config_services)

    def _load_configs(self, config_services):
        """
        loads all required configs of this package.

        this method is intended for overriding by
        subclasses to do custom configurations.

        :param Module config_services: configuration services dependency.
                                       to be able to overcome circular dependency problem,
                                       we should inject configuration services dependency
                                       into this method. because all other packages are
                                       referenced `packaging.base` module in them, so we
                                       can't import `pyrin.configuration.services` in this
                                       module. this is more beautiful in comparison to
                                       importing it inside this method.
        """
        pass

    @property
    def config_defaults(self):
        """
        gets config store default values that should be sent to config parser.

        it is used for interpolation.
        this method is intended to be overridden by subclasses.

        :rtype: dict
        """

        return None



# -*- coding: utf-8 -*-
"""
sentry hooks module.
"""

import pyrin.logging.sentry.services as sentry_services

from pyrin.api.decorators import api_hook
from pyrin.api.hooks import APIHookBase
from pyrin.logging.sentry.integrations import SentryIntegrationsPackage
from pyrin.packaging.decorators import packaging_hook
from pyrin.packaging.hooks import PackagingHookBase


@packaging_hook()
class PackagingHook(PackagingHookBase):
    """
    packaging hook class.
    """

    def package_loaded(self, package_name, **options):
        """
        this method will be called after each application package has been loaded.

        :param str package_name: name of the loaded package.
        """

        # we should call this method as soon as integrations package is
        # loaded, to make sure sentry configures itself on time.
        if package_name == SentryIntegrationsPackage.NAME:
            sentry_services.configure()


@api_hook()
class APIHook(APIHookBase):
    """
    api hook class.
    """

    def exception_occurred(self, error, **options):
        """
        this method will be called when an exception is occurred.

        :param Exception error: exception instance that has been occurred.
        """

        sentry_services.try_report_exception(error, **options)



# -*- coding: utf-8 -*-
"""
sentry integrations logging module.
"""

from sentry_sdk.integrations.logging import LoggingIntegration

import pyrin.configuration.services as config_services

from pyrin.logging.sentry.decorators import sentry_integration
from pyrin.logging.sentry.integrations.base import SentryIntegrationBase


@sentry_integration()
class LoggingSentryIntegration(SentryIntegrationBase):
    """
    logging sentry integration class.
    """

    name = 'logging'

    def _configure(self, **options):
        """
        configures this integration and gets the configured object.

        :rtype: LoggingIntegration
        """

        level = config_services.get('sentry', self.get_name(), 'level')
        event_level = config_services.get('sentry', self.get_name(), 'event_level')
        return LoggingIntegration(level, event_level)



# -*- coding: utf-8 -*-
"""
cli core package.

this package provides facilities to expose
`pyrin` global command after installing pyrin.
"""

from pyrin.packaging.base import Package


class CLICorePackage(Package):
    """
    cli core package class.
    """

    NAME = __name__



# -*- coding: utf-8 -*-
"""
schema manager module.
"""

import pyrin.converters.serializer.services as serializer_services

from pyrin.api.schema import SchemaPackage
from pyrin.core.structs import Manager


class SchemaManager(Manager):
    """
    schema manager class.
    """

    package_class = SchemaPackage

    def get_computed_entity_columns(self, entity, result_schema=None, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        if `result_schema` is not provided, it returns an empty dict.
        note that the result dict should not contain any `BaseEntity` or
        `ROW_RESULT` values, otherwise a max recursion error may occur.

        :param BaseEntity entity: the actual entity to be processed.
        :param ResultSchema result_schema: result schema to be used for computation.

        :rtype: dict
        """

        if result_schema is None:
            return {}

        result = result_schema.get_computed_entity_columns(entity, **options)
        if result is None:
            return {}

        options.update(result_schema=result_schema)
        result = serializer_services.serialize(result, **options)
        return result

    def get_computed_row_columns(self, row, result_schema=None, **options):
        """
        gets a dict containing all computed columns to be added to the result.

        if `result_schema` is not provided, it returns an empty dict.
        note that the result dict should not contain any `BaseEntity` or
        `ROW_RESULT` values, otherwise a max recursion error may occur.

        :param ROW_RESULT row: the actual row result to be processed.
        :param ResultSchema result_schema: result schema to be used for computation.

        :rtype: dict
        """

        if result_schema is None:
            return {}

        result = result_schema.get_computed_row_columns(row, **options)
        if result is None:
            return {}

        options.update(result_schema=result_schema)
        result = serializer_services.serialize(result, **options)
        return result



# -*- coding: utf-8 -*-
"""
application container module.
"""

from threading import Lock

from pyrin.application.exceptions import ApplicationInstanceAlreadySetError


# holds the instance of running application.
__app__ = None
__lock__ = Lock()


def _set_app(app):
    """
    sets the global instance of running application.

    this instance should not be accessed directly from packages outside
    of `pyrin.application` package. outside packages, instead should always
    access application functionality through `pyrin.application.services` module.

    :param Application app: application instance.

    :raises ApplicationInstanceAlreadySetError: application instance already set error.
    """

    global __app__, __lock__

    message = 'Application instance has been already set, it could not be overwritten.'
    if __app__ is not None:
        raise ApplicationInstanceAlreadySetError(message)

    with __lock__:
        if __app__ is not None:
            raise ApplicationInstanceAlreadySetError(message)

        __app__ = app


def _get_app():
    """
    gets the current instance of running application.

    this instance should not be accessed directly from packages outside
    of `pyrin.application` package. outside packages, instead should always
    access application functionality through `pyrin.application.services` module.

    :rtype: Application
    """

    global __app__
    return __app__



# -*- coding: utf-8 -*-
"""
orm sql schema base module.
"""

from sqlalchemy import Column

from pyrin.caching.decorators import cached_property
from pyrin.database.orm.sql.operators.base import CoreColumnOperators


class CoreColumn(Column, CoreColumnOperators):
    """
    core column class.

    all application models columns must be an instance of this class.
    """

    def __init__(self, *args, **kwargs):
        """
        initializes an instance of CoreColumn.

        :param str name: the name of this column as represented in the database.
                         this argument may be the first positional argument, or
                         specified via keyword.

        :param TypeEngine type_: the column's type, indicated using an instance which
                                 if no arguments are required for the type, the class of
                                 the type can be sent as well.
                                 this argument may be the second positional argument, or
                                 specified via keyword.

        :param object *args: additional positional arguments include various
                             `SchemaItem` derived constructs which will be applied
                             as options to the column.

        :keyword bool autoincrement: set up `auto increment` semantics for an
                                     integer primary key column.

        :keyword callable | object default: a scalar, python callable or `ColumnElement`
                                            expression representing the default value
                                            for this column, which will be invoked upon
                                            insert if this column is otherwise not
                                            specified in the values clause of the insert.

        :keyword str doc: optional string that can be used by the ORM or similar
                          to document attributes on the python side.

        :keyword str key: an optional string identifier which will identify this
                          `Column` object on the `Table`.

        :keyword bool index: when `True`, indicates that the column is indexed.

        :keyword dict info: optional data dictionary which will be populated into the
                            `SchemaItem.info` attribute of this object.

        :keyword bool nullable: when set to `False`, will cause the `Not NULL`
                                phrase to be added when generating ddl for the column.

        :keyword callable | object onupdate: a scalar, python callable, or
                                             `ClauseElement` representing a default
                                             value to be applied to the column within update
                                             statements, which will be invoked upon update
                                             if this column is not present in the set
                                             clause of the update.

        :keyword bool primary_key: if `True`, marks this column as a primary key
                                   column. multiple columns can have this flag set to
                                   specify composite primary keys.

        :keyword object server_default: a `FetchedValue` instance, str, unicode
                                        or `text` construct representing the ddl
                                        default value for the column.

        :keyword FetchedValue server_onupdate: a `FetchedValue` instance representing a
                                               database-side default generation function,
                                               such as a trigger. this indicates to sqlalchemy
                                               that a newly generated value will be available
                                               after updates. this construct does not actually
                                               implement any kind of generation function within
                                               the database, which instead must be specified
                                               separately.

        :keyword bool quote: force quoting of this column's name on or off,
                             corresponding to `True` or `False`. when left at its default
                             of `None`, the column identifier will be quoted according to
                             whether the name is case sensitive (identifiers with at least one
                             upper case character are treated as case sensitive), or if it's a
                             reserved word.

        :keyword bool unique: when `True`, indicates that this column contains a
                              unique constraint, or if `index` is `True` as well, indicates
                              that the `index` should be created with the unique flag.

        :keyword bool system: when `True`, indicates this is a system column,
                              that is a column which is automatically made available by the
                              database, and should not be included in the columns list for a
                              `create table` statement.

        :keyword str comment: optional string that will render an sql comment
                              on table creation.

        :keyword bool allow_read: specifies that the column should be
                                  included in entity to dict conversion.
                                  defaults to True if not provided.

        :keyword bool allow_write: specifies that the column should be
                                   populated on conversion from dict.
                                   defaults to True if not provided.
        """

        self.allow_read = kwargs.pop('allow_read', True)
        self.allow_write = kwargs.pop('allow_write', True)

        super().__init__(*args, **kwargs)

    @cached_property
    def fullname(self):
        """
        gets the column's fullname.

        fullname is made up of `table_fullname.column_name`.
        if the column has no table, it only returns the `column_name`.

        :rtype: str
        """

        if self.table is None:
            return self._real_name()

        for single_column in self.table.columns:
            if single_column.name == self.name:
                for base_column in single_column.base_columns:
                    return '{table}.{column}'.format(table=base_column.table.fullname,
                                                     column=self._real_name())

    @property
    def is_foreign_key(self):
        """
        gets a value indicating that this column is a foreign key.

        :rtype: bool
        """

        return len(self.foreign_keys) > 0

    def _real_name(self):
        """
        gets the column's real name.

        :rtype: str
        """

        for column in self.base_columns:
            return column.name



# -*- coding: utf-8 -*-
"""
database interface module.
"""

from abc import abstractmethod
from threading import Lock

from pyrin.core.structs import CoreObject, MultiSingletonMeta
from pyrin.core.exceptions import CoreNotImplementedError


class SessionFactorySingletonMeta(MultiSingletonMeta):
    """
    session factory singleton meta class.
    this is a thread-safe implementation of singleton.
    """

    _instances = dict()
    _lock = Lock()


class AbstractSessionFactoryBase(CoreObject, metaclass=SessionFactorySingletonMeta):
    """
    abstract session factory base class.
    """

    @abstractmethod
    def create_session_factory(self, engine):
        """
        creates a database session factory and binds it to given engine.

        :param Engine engine: database engine.

        :raises CoreNotImplementedError: core not implemented error.

        :returns: database session
        :rtype: Session
        """

        raise CoreNotImplementedError()

    @property
    @abstractmethod
    def request_bounded(self):
        """
        gets a value indicating that this session factory type should be bounded into request.

        :raises CoreNotImplementedError: core not implemented error.

        :rtype: bool
        """

        raise CoreNotImplementedError()



# -*- coding: utf-8 -*-
"""
babel handlers params module.
"""

import os

import pyrin.configuration.services as config_services
import pyrin.application.services as application_services
import pyrin.globalization.locale.babel.services as babel_services

from pyrin.globalization.locale.babel.interface import BabelCLIParamBase
from pyrin.cli.arguments import KeywordArgument, BooleanArgument, \
    PositionalArgument, CompositeKeywordArgument


class DomainParam(KeywordArgument, BabelCLIParamBase):
    """
    domain param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DomainParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to `domain` value form babel
                               config store if not provided.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'domain')

        super().__init__('domain', '--domain', default=default)


class CompileDomainsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    compile domains param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CompileDomainsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to `compile_domains` value form
                               babel config store if not provided.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'compile_domains')

        super().__init__('domain', '--domain', default=default)


class InputTemplateFileParam(KeywordArgument, BabelCLIParamBase):
    """
    input template file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of InputTemplateFileParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `locale_path/template_file`
                               where `template_file` comes from
                               babel config store.
        """

        if default is None:
            default = os.path.join(application_services.get_locale_path(),
                                   config_services.get('babel', 'arguments', 'template_file'))

        super().__init__('input_file', '--input-file', default=default)


class OutputDirectoryParam(KeywordArgument, BabelCLIParamBase):
    """
    output directory param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputDirectoryParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `locale_path` of application.
        """

        if default is None:
            default = application_services.get_locale_path()

        super().__init__('output_dir', '--output-dir', default=default)


class OutputTemplateFileParam(KeywordArgument, BabelCLIParamBase):
    """
    output template file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputTemplateFileParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `locale_path/template_file`
                               where `template_file` comes from
                               babel config store.
        """

        if default is None:
            default = os.path.join(application_services.get_locale_path(),
                                   config_services.get('babel', 'arguments', 'template_file'))

        super().__init__('output_file', '--output-file', default=default)


class OutputFileParam(KeywordArgument, BabelCLIParamBase):
    """
    output file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OutputFileParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to None.
        """

        super().__init__('output_file', '--output-file', default=default)


class OmitHeaderParam(BooleanArgument, BabelCLIParamBase):
    """
    omit header param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of OmitHeaderParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `omit_header`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'omit_header')

        super().__init__('omit_header', '--omit-header', default=default)


class LocaleParam(KeywordArgument, BabelCLIParamBase):
    """
    locale param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of LocaleParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to None.
        """

        super().__init__('locale', '--locale', default=default)


class WidthParam(KeywordArgument, BabelCLIParamBase):
    """
    width param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of WidthParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `width`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'width')

        super().__init__('width', '--width', default=default)


class NoWrapParam(BooleanArgument, BabelCLIParamBase):
    """
    no wrap param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoWrapParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `no_wrap`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_wrap')

        super().__init__('no_wrap', '--no-wrap', default=default)


class IgnoreObsoleteParam(BooleanArgument, BabelCLIParamBase):
    """
    ignore obsolete param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of IgnoreObsoleteParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `ignore_obsolete`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'ignore_obsolete')

        super().__init__('ignore_obsolete', '--ignore-obsolete', default=default)


class NoFuzzyMatchingParam(BooleanArgument, BabelCLIParamBase):
    """
    no fuzzy matching param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoFuzzyMatchingParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `no_fuzzy_matching`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_fuzzy_matching')

        super().__init__('no_fuzzy_matching', '--no-fuzzy-matching', default=default)


class UpdateHeaderCommentParam(BooleanArgument, BabelCLIParamBase):
    """
    update header comment param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of UpdateHeaderCommentParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `update_header_comment`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'update_header_comment')

        super().__init__('update_header_comment', '--update-header-comment', default=default)


class PreviousParam(BooleanArgument, BabelCLIParamBase):
    """
    previous param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of PreviousParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `previous`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'previous')

        super().__init__('previous', '--previous', default=default)


class CharsetParam(KeywordArgument, BabelCLIParamBase):
    """
    charset param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CharsetParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `charset`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'charset')

        super().__init__('charset', '--charset', default=default)


class KeywordsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    keywords param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of KeywordsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `keywords`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'keywords')

        super().__init__('keywords', '--keywords', default=default)


class NoDefaultKeywordsParam(BooleanArgument, BabelCLIParamBase):
    """
    no default keywords param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoDefaultKeywordsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `no_default_keywords`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_default_keywords')

        super().__init__('no_default_keywords', '--no-default-keywords', default=default)


class MappingParam(KeywordArgument, BabelCLIParamBase):
    """
    mapping param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MappingParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               path to `babel.mappings.ini`.
        """

        if default is None:
            package = babel_services.get_package_class()
            default = config_services.get_file_path(package.EXTRA_CONFIG_STORE_NAMES[0])

        super().__init__('mapping', '--mapping', default=default)


class NoLocationParam(BooleanArgument, BabelCLIParamBase):
    """
    no location param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of NoLocationParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `no_location`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'no_location')

        super().__init__('no_location', '--no-location', default=default)


class AddLocationParam(KeywordArgument, BabelCLIParamBase):
    """
    add location param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AddLocationParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `add_location`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'add_location')

        super().__init__('add_location', '--add-location', default=default)


class SortOutputParam(BooleanArgument, BabelCLIParamBase):
    """
    sort output param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SortOutputParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `sort_output`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'sort_output')

        super().__init__('sort_output', '--sort-output', default=default)


class SortByFileParam(BooleanArgument, BabelCLIParamBase):
    """
    sort by file param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of SortByFileParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `sort_by_file`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'sort_by_file')

        super().__init__('sort_by_file', '--sort-by-file', default=default)


class MSGIDBugsAddressParam(KeywordArgument, BabelCLIParamBase):
    """
    msgid bugs address param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of MSGIDBugsAddressParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `msgid_bugs_address`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'msgid_bugs_address')

        super().__init__('msgid_bugs_address', '--msgid-bugs-address', default=default)


class CopyrightHolderParam(KeywordArgument, BabelCLIParamBase):
    """
    copyright holder param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of CopyrightHolderParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `copyright_holder`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'copyright_holder')

        super().__init__('copyright_holder', '--copyright-holder', default=default)


class ProjectParam(KeywordArgument, BabelCLIParamBase):
    """
    project param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of ProjectParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `project`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'project')

        super().__init__('project', '--project', default=default)


class VersionParam(KeywordArgument, BabelCLIParamBase):
    """
    version param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of VersionParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `version`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'version')

        super().__init__('version', '--version', default=default)


class AddCommentsParam(CompositeKeywordArgument, BabelCLIParamBase):
    """
    add comments param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of AddCommentsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `add_comments`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'add_comments')

        super().__init__('add_comments', '--add-comments', default=default)


class StripCommentsParam(BooleanArgument, BabelCLIParamBase):
    """
    strip comments param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of StripCommentsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `strip_comments`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'strip_comments')

        super().__init__('strip_comments', '--strip-comments', default=default)


class InputPathsParam(PositionalArgument, BabelCLIParamBase):
    """
    input paths param class.
    """

    def __init__(self, index=None, default=None, **options):
        """
        initializes an instance of InputPathsParam.

        :param int index: zero based index of this param in cli command inputs.
                          defaults to 0 if not provided.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               defaults to both application and pyrin
                               main package paths if not provided.

        :keyword bool validate_index: specifies that index of this argument
                                      must be validated. it could be helpful
                                      to set this to False when there are multiple
                                      arguments with the same index that will appear
                                      in different situations.
                                      defaults to True if not provided.
        """

        if index is None:
            index = 0

        if default is None:
            default = self._get_paths(include_application=True, include_pyrin=True)

        super().__init__('input_paths', index, default=default, **options)

    def _process_inputs(self, **options):
        """
        processes the inputs given to this handler and returns them.

        :rtype: dict
        """

        include_application = options.get('include_app', False)
        include_pyrin = options.get('include_pyrin', False)
        paths = self._get_paths(include_application, include_pyrin)

        if len(paths) <= 0:
            paths = None

        options.update(input_paths=paths)
        return super()._process_inputs(**options)

    def _get_paths(self, include_application=True, include_pyrin=True):
        """
        gets application and pyrin main package paths based on given inputs.

        :param bool include_application: include application main package path.
                                         defaults to True if not provided.

        :param bool include_pyrin: include pyrin main package path.
                                   defaults to True if not provided.

        :rtype: list[str]
        """

        paths = []
        if include_application is True:
            paths.append(application_services.get_application_main_package_path())

        if include_pyrin is True:
            paths.append(application_services.get_pyrin_main_package_path())

        return paths


class DirectoryParam(KeywordArgument, BabelCLIParamBase):
    """
    directory param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of DirectoryParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to
                               `locale_path` of application.
        """

        if default is None:
            default = application_services.get_locale_path()

        super().__init__('directory', '--directory', default=default)


class UseFuzzyParam(BooleanArgument, BabelCLIParamBase):
    """
    use fuzzy param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of UseFuzzyParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to `use_fuzzy`
                               value from babel config store.
        """

        if default is None:
            default = config_services.get('babel', 'arguments', 'use_fuzzy')

        super().__init__('use_fuzzy', '--use-fuzzy', default=default)


class StatisticsParam(BooleanArgument, BabelCLIParamBase):
    """
    statistics param class.
    """

    def __init__(self, default=None):
        """
        initializes an instance of StatisticsParam.

        :param object default: default value to be emitted to
                               cli if this param is not available.
                               if set to None, this param will not
                               be emitted at all.
                               if not provided, defaults to None.
        """

        super().__init__('statistics', '--statistics', default=default)



# -*- coding: utf-8 -*-
"""
audit decorators module.
"""

import pyrin.audit.services as audit_services


def audit_hook():
    """
    decorator to register an audit hook.

    :raises InvalidAuditHookTypeError: invalid audit hook type error.

    :returns: audit hook class.
    :rtype: type
    """

    def decorator(cls):
        """
        decorates the given class and registers an instance
        of it into available audit hooks.

        :param type cls: audit hook class.

        :returns: audit hook class.
        :rtype: type
        """

        instance = cls()
        audit_services.register_hook(instance)

        return cls

    return decorator



# -*- coding: utf-8 -*-
"""
token component module.
"""

from pyrin.application.decorators import component
from pyrin.security.token import TokenPackage
from pyrin.security.token.manager import TokenManager
from pyrin.application.structs import Component


@component(TokenPackage.COMPONENT_NAME)
class TokenComponent(Component, TokenManager):
    """
    token component class.
    """
    pass



# -*- coding: utf-8 -*-
"""
mimetype package.
"""

from pyrin.packaging.base import Package


class MIMETypePackage(Package):
    """
    mimetype package class.
    """

    NAME = __name__
    COMPONENT_NAME = 'processor.mimetype.component'



# -*- coding: utf-8 -*-
"""
pyrin setup module.
"""

import io
import re

from setuptools import find_namespace_packages
from setuptools import setup


with io.open('README.md', 'rt', encoding='utf8') as readme_file:
    README = readme_file.read()

with io.open('src/pyrin/__init__.py', 'rt', encoding='utf8') as version_file:
    VERSION = re.search(r"__version__ = '(.*?)'", version_file.read()).group(1)

PACKAGES = [
    'aniso8601>=8.0.0',
    'bcrypt>=3.1.7',
    'pytz>=2019.3',
    'Flask>=1.1.1',
    'PyJWT>=2.0.1',
    'SQLAlchemy>=1.3.13, <1.4',
    'colorama>=0.4.3',
    'python-dotenv>=0.10.5',
    'cryptography>=2.8',
    'flask-babel>=1.0.0',
    'babel>=2.8.0',
    'alembic>=1.4.0',
    'fire>=0.2.1',
    'sqlparse>=0.3.0',
]

TEST_PACKAGES = PACKAGES + [
    'pytest',
    'pytest-cov',
    'pygments',
]

DOC_PACKAGES = [
    'sphinx',
    'sphinxcontrib-log-cabinet',
    'sphinx-issues',
]

MEMCACHED_PACKAGES = [
    'pymemcache>=3.3.0',
]

SENTRY_PACKAGES = [
    'sentry-sdk>=0.17.4',
    'blinker>=1.4',
]

CELERY_PACKAGES = [
    'celery>=4.4.7',
]

REDIS_PACKAGES = [
    'redis>=3.5.3',
]

setup(
    name='pyrin',
    version=VERSION,
    url='https://github.com/mononobi/pyrin',
    project_urls={
        # 'Documentation': '',
        'Code': 'https://github.com/mononobi/pyrin',
        'Issue tracker': 'https://github.com/mononobi/pyrin/issues',
    },
    license='GPL-3.0',
    author='Mohamad Nobakht',
    author_email='mohamadnobakht@gmail.com',
    maintainer='Mohamad Nobakht',
    maintainer_email='mohamadnobakht@gmail.com',
    description='A rich platform-independent application framework '
                'to build apps using Flask on top of it.',
    long_description=README,
    long_description_content_type='text/markdown',
    keywords=('flask python sqlalchemy pyrin core alembic multi-database '
              'application-framework rest-api dependency-injection ioc'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_namespace_packages('src', exclude=('tests', 'tests.*')),
    package_dir={'': 'src'},
    package_data={'': ['*']},
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=PACKAGES,
    extras_require={
        'tests': TEST_PACKAGES,
        'docs': DOC_PACKAGES,
        'memcached': MEMCACHED_PACKAGES,
        'sentry': SENTRY_PACKAGES,
        'celery': CELERY_PACKAGES,
        'redis': REDIS_PACKAGES,
    },
    entry_points={"console_scripts": ["pyrin = pyrin.cli.core.command:main"]},
)



# -*- coding: utf-8 -*-
"""
utils misc module.
"""

from pyrin.core.globals import LIST_TYPES
from pyrin.core.exceptions import CoreAttributeError


def set_attributes(instance, **kwargs):
    """
    sets the provided keyword arguments as attributes in given object instance.

    :param CoreObject instance: instance to set attributes for.

    :returns: instance with updated attributes.
    :rtype: CoreObject
    """

    if instance is None:
        return instance

    for name, value in kwargs.items():
        if hasattr(instance, name):
            delattr(instance, name)

        setattr(instance, name, value)

    return instance


def extract_attributes(instance, *attributes):
    """
    extracts all attributes with given names from provided object instance.

    :param CoreObject instance: instance to extract attributes from.
    :param attributes: list of all attribute names that should be extracted.

    :raises CoreAttributeError: core attribute error.

    :returns: dict[str name: object value]
    :rtype: dict
    """

    result = dict()
    if instance is None:
        return result

    for name in attributes:
        if not hasattr(instance, name):
            raise CoreAttributeError('Object [{object}] has no attribute [{name}].'
                                     .format(object=str(instance), name=name))
        result[name] = getattr(instance, name)

    return result


def make_iterable(values, collection=None):
    """
    converts the provided values to iterable.

    it returns a collection of values using the given collection type.

    :param object | list[object] | tuple[object] | set[object] values: value or values to make
                                                                       iterable. if the values
                                                                       are iterable, it just
                                                                       converts the collection
                                                                       to given type.

    :param type[list | tuple | set] collection: collection type to use.
                                                defaults to list if not provided.

    :rtype: list | tuple | set
    """

    if collection is None:
        collection = list

    if values is None:
        return collection()

    if not isinstance(values, LIST_TYPES):
        values = (values,)

    return collection(values)


def try_get_fully_qualified_name(some_object):
    """
    tries to get the fully qualified name of given object.

    it tries to return `__module__.__name__` for given object.
    for example: `pyrin.api.services.create_route`.
    but if it fails to get any of those, it returns the `__str__` for that object.

    :param object some_object: object to get its fully qualified name.

    :rtype: str
    """

    module = None
    name = None
    try:
        module = some_object.__module__
        if module == '' or module.isspace():
            module = None
    except AttributeError:
        module = None

    try:
        name = some_object.__name__
        if name == '' or name.isspace():
            name = None
    except AttributeError:
        name = None

    if module is not None and name is not None:
        return '{module}.{name}'.format(module=module, name=name)

    return str(some_object)


def iterate_items(collection, *args, **kwargs):
    """
    iterates over the items of given object.

    :param type collection: collection type to be used to
                            iterate over given object's items.

    :rtype: iterator[tuple[object, object]]
    """

    return iter(collection.items(*args, **kwargs))



# -*- coding: utf-8 -*-
"""
string normalizer manager module.
"""

from collections import OrderedDict

import pyrin.utils.dictionary as dict_utils

from pyrin.core.structs import Context, Manager
from pyrin.utilities.string.normalizer import StringNormalizerPackage
from pyrin.utilities.string.normalizer.enumerations import NormalizerEnum
from pyrin.utilities.string.normalizer.exceptions import StringNormalizerDoesNotExistError
from pyrin.utilities.string.normalizer.interface import AbstractStringNormalizerBase
from pyrin.utils.custom_print import print_warning
from pyrin.utilities.string.normalizer.exceptions import InvalidStringNormalizerTypeError, \
    DuplicatedStringNormalizerError


class StringNormalizerManager(Manager):
    """
    string normalizer manager class.
    """

    package_class = StringNormalizerPackage

    def __init__(self):
        """
        initializes an instance of StringNormalizerManager.
        """

        super().__init__()

        # a dictionary containing information of registered normalizers.
        # example: dict(str name: AbstractStringNormalizerBase instance)
        self._normalizers = Context()

        # a dictionary containing name of normalizers and their priority.
        # values are ordered from max to min priority.
        # example: dict(str name: int priority)
        self._priorities = OrderedDict()

    def normalize(self, value, *normalizers, **options):
        """
        normalizes the given value.

        :param str value: value to be normalized.

        :param str normalizers: normalizer names to be used.
                                they will be used in the order of their appearance.
                                if not provided, all normalizers will be used with
                                the order of their priority attribute.

        :keyword list[str] filters: list of items to be removed from string.
                                    defaults to None. it will only be used
                                    for `filter` normalizer.

        :keyword bool ignore_case: remove `filters` from string in case-insensitive
                                   way. defaults to True if not provided.

        :keyword bool strip: strip spaces from both ends of string on each
                             normalization step. defaults to True if not provided.

        :raises StringNormalizerDoesNotExistError: string normalizer does not exist error.

        :returns: normalized value.
        :rtype: str
        """

        if value in (None, ''):
            return ''

        if normalizers is None or len(normalizers) <= 0:
            normalizers = self._priorities.keys()

        for name in normalizers:
            normalizer = self.get_normalizer(name)
            value = normalizer.normalize(value, **options)

            if value == '':
                break

        return value

    def register_normalizer(self, instance, **options):
        """
        registers a new string normalizer or replaces the existing one.

        if `replace=True` is provided. otherwise, it raises an error
        on adding a normalizer which is already registered.

        :param AbstractStringNormalizerBase instance: normalizer to be registered.
                                                      it must be an instance of
                                                      AbstractStringNormalizerBase.

        :keyword bool replace: specifies that if there is another registered
                               normalizer with the same name, replace it with
                               the new one, otherwise raise an error.
                               defaults to False.

        :raises InvalidStringNormalizerTypeError: invalid string normalizer type error.
        :raises DuplicatedStringNormalizerError: duplicated string normalizer error.
        """

        if not isinstance(instance, AbstractStringNormalizerBase):
            raise InvalidStringNormalizerTypeError('Input parameter [{instance}] is '
                                                   'not an instance of [{base}].'
                                                   .format(instance=instance,
                                                           base=AbstractStringNormalizerBase))

        if instance.get_name() in self._normalizers:
            old_instance = self._normalizers.get(instance.get_name())
            replace = options.get('replace', False)
            if replace is not True:
                raise DuplicatedStringNormalizerError('There is another registered '
                                                      'string normalizer with name [{name}] '
                                                      'but "replace" option is not set, so '
                                                      'normalizer [{instance}] could not '
                                                      'be registered.'
                                                      .format(name=instance.get_name(),
                                                              instance=instance))

            print_warning('String normalizer [{old_instance}] is going '
                          'to be replaced by [{new_instance}].'
                          .format(old_instance=old_instance,
                                  new_instance=instance))

        self._normalizers[instance.get_name()] = instance
        self._priorities[instance.get_name()] = instance.priority
        self._priorities = OrderedDict(dict_utils.sort_by_value(self._priorities,
                                                                reverse=True))

    def get_normalizer(self, name, **options):
        """
        gets the string normalizer with given name.

        it raises an error if normalizer does not exist.

        :raises StringNormalizerDoesNotExistError: string normalizer does not exist error.

        :rtype: AbstractStringNormalizerBase
        """

        if name in self._normalizers:
            return self._normalizers.get(name)

        raise StringNormalizerDoesNotExistError('String normalizer with name [{name}] '
                                                'does not exist.'.format(name=name))

    def space(self, value, **options):
        """
        remove all spaces from the string.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.SPACE)
        return normalizer.normalize(value, **options)

    def duplicate_space(self, value, **options):
        """
        replace all duplicate spaces with a single one.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.DUPLICATE_SPACE)
        return normalizer.normalize(value, **options)

    def lowercase(self, value, **options):
        """
        lowercase the string.

        :param str value: value to be lower-cased.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.LOWERCASE)
        return normalizer.normalize(value, **options)

    def uppercase(self, value, **options):
        """
        uppercase the string.

        :param str value: value to be upper-cased.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.UPPERCASE)
        return normalizer.normalize(value, **options)

    def title_case(self, value, **options):
        """
        title case the string.

        :param str value: value to be title-cased.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.TITLE_CASE)
        return normalizer.normalize(value, **options)

    def filter(self, value, **options):
        """
        remove filters from the string.

        :param str value: value to be filtered.

        :keyword list[str] filters: list of items to be removed from value.

        :keyword bool ignore_case: remove `filters` from string in case-insensitive
                                   way. defaults to True if not provided.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.FILTER)
        return normalizer.normalize(value, **options)

    def persian_sign(self, value, **options):
        """
        remove common persian signs from the string.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.PERSIAN_SIGN)
        return normalizer.normalize(value, **options)

    def latin_sign(self, value, **options):
        """
        remove common latin signs from the string.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.LATIN_SIGN)
        return normalizer.normalize(value, **options)

    def persian_number(self, value, **options):
        """
        replace persian numbers with latin numbers.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.PERSIAN_NUMBER)
        return normalizer.normalize(value, **options)

    def arabic_number(self, value, **options):
        """
        replace arabic numbers with latin numbers.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.ARABIC_NUMBER)
        return normalizer.normalize(value, **options)

    def persian_letter(self, value, **options):
        """
        normalize persian letters.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.PERSIAN_LETTER)
        return normalizer.normalize(value, **options)

    def latin_letter(self, value, **options):
        """
        normalize latin letters.

        :param str value: value to be normalized.

        :rtype: str
        """

        normalizer = self.get_normalizer(NormalizerEnum.LATIN_LETTER)
        return normalizer.normalize(value, **options)



# -*- coding: utf-8 -*-
"""
permission services module.
"""

from pyrin.application.services import get_component
from pyrin.database.transaction.decorators import atomic
from pyrin.security.permission import PermissionPackage


def register_permission(instance, **options):
    """
    registers the given permission.

    :param PermissionBase instance: permission instance to be registered.

    :raises InvalidPermissionTypeError: invalid permission type error.
    :raises DuplicatedPermissionError: duplicated permission error.
    """

    return get_component(PermissionPackage.COMPONENT_NAME).register_permission(instance,
                                                                               **options)


def get_permissions(**options):
    """
    gets all registered permissions.

    :rtype: list[PermissionBase]
    """

    return get_component(PermissionPackage.COMPONENT_NAME).get_permissions(**options)


@atomic
def synchronize_all(**options):
    """
    synchronizes all permissions with database.

    it creates or updates the available permissions.
    """

    return get_component(PermissionPackage.COMPONENT_NAME).synchronize_all(**options)



# -*- coding: utf-8 -*-
"""
caching services module.
"""

from pyrin.application.services import get_component
from pyrin.caching import CachingPackage
from pyrin.database.transaction.decorators import atomic


def register_cache(instance, **options):
    """
    registers a new cache or replaces the existing one.

    if `replace=True` is provided. otherwise, it raises an error
    on adding a cache which is already registered.

    :param AbstractCache instance: cache instance to be registered.

    :keyword bool replace: specifies that if there is another registered
                           cache with the same name, replace it with the
                           new one, otherwise raise an error. defaults to False.

    :raises InvalidCacheTypeError: invalid cache type error.
    :raises DuplicatedCacheError: duplicated cache error.
    """

    get_component(CachingPackage.COMPONENT_NAME).register_cache(instance, **options)


def get_cache(name):
    """
    gets the registered cache with given name.

    it raises an error if no cache found for given name.

    :param str name: name of cache to be get.

    :raises CacheNotFoundError: cache not found error.

    :rtype: AbstractCache
    """

    return get_component(CachingPackage.COMPONENT_NAME).get_cache(name)


def contains(name, key):
    """
    gets a value indicating that given key is existed in the cached items of given cache.

    :param str name: name of the cache.
    :param object key: key to be checked for existence.

    :raises CacheNotFoundError: cache not found error.

    :rtype: bool
    """

    return get_component(CachingPackage.COMPONENT_NAME).contains(name, key)


def pop(name, key, default=None):
    """
    pops the given key from cached items of given cache and returns its value.

    if key does not exist, it returns None or the specified default value.

    :param str name: name of the cache.
    :param object key: key to get its value.
    :param object default: value to be returned if key is not present.

    :raises CacheNotFoundError: cache not found error.

    :returns: object
    """

    return get_component(CachingPackage.COMPONENT_NAME).pop(name, key, default=default)


def remove(name, key, **options):
    """
    removes the given key from cached items of given cache.

    it does nothing if the key is not present in the cache.

    :param str name: name of the cache.
    :param object key: key to be removed.

    :raises CacheNotFoundError: cache not found error.
    """

    get_component(CachingPackage.COMPONENT_NAME).remove(name, key, **options)


def clear(name):
    """
    clears a cache with given name.

    :param str name: cache name to be cleared.

    :raises CacheNotFoundError: cache not found error.
    """

    get_component(CachingPackage.COMPONENT_NAME).clear(name)


def set(name, key, value, **options):
    """
    sets a new value into given cache.

    :param str name: cache name.
    :param object key: hashable key of the cache to be registered.
    :param object value: value to be cached.

    :keyword int expire: expire time for given key in milliseconds.
                         if not provided, will be get from `caching` config store.
                         this value is only used in complex and remote caches.

    :keyword bool refreshable: specifies that cached item's expire time must be
                               extended on each hit. if not provided, will be get
                               from `caching` config store.
                               this value is only used in complex caches.

    :raises CacheNotFoundError: cache not found error.
    """

    get_component(CachingPackage.COMPONENT_NAME).set(name, key, value, **options)


def get(name, key, default=None, **options):
    """
    gets the value from given cache.

    if key does not exist, it returns None or the specified default value.

    :param str name: cache name.
    :param object key: hashable key to get its value from cache.
    :param object default: value to be returned if key is not present.

    :raises CacheNotFoundError: cache not found error.

    :returns: object
    """

    return get_component(CachingPackage.COMPONENT_NAME).get(name, key,
                                                            default=default, **options)


def try_set(name, value, func, *extra_keys, **options):
    """
    sets a new value into given cache.

    this method will generate cache key from given inputs.
    if the provided values are not hashable, this method won't raise
    an error and logs it silently.

    :param str name: cache name.
    :param object value: value to be cached.
    :param function func: function to cache its result.

    :param type | object parent: parent class or instance of given function.
                                 this should only passed to simple permanent caches.

    :param tuple inputs: function positional arguments.
                         this should only passed to extended and complex caches.

    :param dict kw_inputs: function keyword arguments.
                           this should only passed to extended, complex and remote caches.

    :param object extra_keys: extra arguments to generate key from.
                              this could be used in custom caches.

    :keyword bool consider_user: specifies that current user must also be
                                 included in cache key. if not provided, will
                                 be get from `caching` config store.
                                 this value is only used in complex, extended
                                 and remote caches.

    :keyword int expire: expire time for given key in milliseconds.
                         if not provided, will be get from `caching` config store.
                         this value is only used in complex and remote caches.

    :keyword bool refreshable: specifies that cached item's expire time must be
                               extended on each hit. if not provided, will be get
                               from `caching` config store.
                               this value is only used in complex caches.

    :raises CacheNotFoundError: cache not found error.
    """

    get_component(CachingPackage.COMPONENT_NAME).try_set(name, value, func,
                                                         *extra_keys, **options)


def try_get(name, func, *extra_keys, default=None, **options):
    """
    gets the value from given cache.

    this method will generate cache key from given inputs.
    if key does not exist, it returns None or the specified default value.
    if the provided values are not hashable, this method won't raise
    an error and logs it silently.

    :param str name: cache name.
    :param function func: function to to get its result.

    :param type | object parent: parent class or instance of given function.
                                 this should only passed to simple permanent caches.

    :param tuple inputs: function positional arguments.
                         this should only passed to extended and complex caches.

    :param dict kw_inputs: function keyword arguments.
                           this should only passed to extended, complex and remote caches.

    :param object extra_keys: extra arguments to generate key from.
                              this could be used in custom caches.

    :param object default: value to be returned if key is not present.

    :keyword bool consider_user: specifies that current user must also be
                                 included in cache key. if not provided, will
                                 be get from `caching` config store.
                                 this value is only used in complex, extended
                                 and remote caches.

    :raises CacheNotFoundError: cache not found error.

    :returns: object
    """

    return get_component(CachingPackage.COMPONENT_NAME).try_get(name, func, *extra_keys,
                                                                default=default, **options)


def generate_key(name, func, *extra_keys, **options):
    """
    generates a cache key from given inputs for the given cache.

    :param str name: cache name.
    :param function func: function to to be cached.

    :param type | object parent: parent class or instance of given function.
                                 this should only passed to simple permanent caches.

    :param tuple inputs: function positional arguments.
                         this should only passed to extended and complex caches.

    :param dict kw_inputs: function keyword arguments.
                           this should only passed to extended, complex and remote caches.

    :param object extra_keys: extra arguments to generate key from.
                              this could be used in custom caches.

    :keyword bool consider_user: specifies that current user must also be
                                 included in cache key. if not provided, will
                                 be get from `caching` config store.
                                 this value is only used in complex, extended
                                 and remote caches.

    :raises CacheNotFoundError: cache not found error.

    :returns: hash of generated key.
    :rtype: int
    """

    return get_component(CachingPackage.COMPONENT_NAME).generate_key(name, func,
                                                                     *extra_keys, **options)


def exists(name):
    """
    returns a value indicating that a cache with the given name existed.

    :param str name: cache name.

    :rtype: bool
    """

    return get_component(CachingPackage.COMPONENT_NAME).exists(name)


def get_cache_names():
    """
    gets all available cache names.

    :rtype: list[str]
    """

    return get_component(CachingPackage.COMPONENT_NAME).get_cache_names()


def get_stats(name):
    """
    gets statistic info of given cache.

    :param str name: cache name to get its info.

    :raises CacheNotFoundError: cache not found error.

    :rtype: dict
    """

    return get_component(CachingPackage.COMPONENT_NAME).get_stats(name)


def get_all_stats():
    """
    gets statistic info of all caches.

    :rtype: dict
    """

    return get_component(CachingPackage.COMPONENT_NAME).get_all_stats()


def persist(name, **options):
    """
    saves cached items of given cache into database.

    :param str name: cache name to be persisted.

    :keyword bool clear: clear all caches after persisting them into database.
                         defaults to False if not provided.

    :raises CacheNotFoundError: cache not found error.
    :raises CacheIsNotPersistentError: cache is not persistent error.
    """

    return get_component(CachingPackage.COMPONENT_NAME).persist(name, **options)


@atomic
def persist_all(**options):
    """
    saves cached items of all persistent caches into database.

    :keyword bool clear: clear all caches after persisting them into database.
                         defaults to False if not provided.
    """

    return get_component(CachingPackage.COMPONENT_NAME).persist_all(**options)


def load(name, **options):
    """
    loads cached items of given cache from database.

    :param str name: cache name to be loaded.

    :raises CacheNotFoundError: cache not found error.
    :raises CacheIsNotPersistentError: cache is not persistent error.
    """

    return get_component(CachingPackage.COMPONENT_NAME).load(name, **options)


@atomic
def load_all(**options):
    """
    loads cached items of all persistent caches from database.
    """

    return get_component(CachingPackage.COMPONENT_NAME).load_all(**options)


def clear_required_caches():
    """
    clears all caches that are required.

    normally, you should never call this method manually. but it is
    implemented to be used for clearing extended and complex caches after
    application has been fully loaded. to enforce that valid results are
    cached based on loaded packages.
    """

    return get_component(CachingPackage.COMPONENT_NAME).clear_required_caches()



# -*- coding: utf-8 -*-
"""
alembic handlers base module.
"""

from pyrin.cli.params import VerboseParam
from pyrin.database.migration.alembic.interface import AlembicCLIHandlerBase
from pyrin.database.migration.alembic.handlers.params import SQLParam, TagParam, RevisionParam


class AlembicUpgradeDowngradeCLIHandlerBase(AlembicCLIHandlerBase):
    """
    alembic upgrade downgrade cli handler base class.

    all alembic upgrade or downgrade cli handlers must be subclassed from this.
    """

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.extend([SQLParam(), TagParam(), RevisionParam()])
        return super()._inject_params(params)


class AlembicReportingCLIHandlerBase(AlembicCLIHandlerBase):
    """
    alembic reporting cli handler base class.

    all alembic reporting cli handlers must be subclassed from this.
    """

    def _inject_params(self, params):
        """
        injects all the params of current handler into given list.

        :param list[CLIParamBase] params: list of all params.
        """

        params.append(VerboseParam())
        return super()._inject_params(params)



# -*- coding: utf-8 -*-
"""
application structs module.
"""

from threading import Lock

from pyrin.core.structs import Context, CoreObject, UniqueSingletonMeta
from pyrin.settings.static import DEFAULT_COMPONENT_KEY
from pyrin.application.exceptions import ComponentAttributeError, InvalidComponentNameError


class ApplicationContext(Context):
    """
    context class to hold application contextual data.
    """

    attribute_error_message = 'Property [{name}] not found in application context.'


class ApplicationComponent(ApplicationContext):
    """
    context class to hold application components.
    """

    attribute_error = ComponentAttributeError
    attribute_error_message = 'Component [{name}] is not available in application components.'


class Component(CoreObject):
    """
    base component class.

    all component classes must inherit from this class and their respective manager class.
    """

    def __init__(self, component_name, **options):
        """
        initializes an instance of Component.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.
        """

        super().__init__()

        # component id is a tuple[str, object] and should be unique for each
        # instance unless it's intended to replace an already existing one.
        self._component_id = self.make_component_id(component_name, **options)

    def get_id(self):
        """
        gets the component id of this instance.

        :rtype: tuple[str, object]
        """

        return self._component_id

    @staticmethod
    def make_component_id(component_name, **options):
        """
        makes a component id based on input values and returns it.

        :param str component_name: component name.

        :keyword object component_custom_key: component custom key.

        :raises InvalidComponentNameError: invalid component name.

        :rtype: tuple[str, object]
        """

        if component_name is None or len(component_name.strip()) <= 0:
            raise InvalidComponentNameError('Component name should not be None.')

        component_custom_key = options.get('component_custom_key', DEFAULT_COMPONENT_KEY)
        return component_name, component_custom_key


class ApplicationSingletonMeta(UniqueSingletonMeta):
    """
    application singleton meta class.

    this is a thread-safe implementation of singleton.
    """

    instance = None
    _lock = Lock()



# -*- coding: utf-8 -*-
"""
database enumerations module.
"""

from pyrin.core.enumerations import CoreEnum


class FKRestrictionEnum(CoreEnum):
    """
    fk restriction enum.
    """

    CASCADE = 'CASCADE'
    DELETE = 'DELETE'
    RESTRICT = 'RESTRICT'



# -*- coding: utf-8 -*-
"""
utils function module.
"""

import inspect

from inspect import Parameter


def get_doc(func, include_returns=True):
    """
    gets docstring of given function.

    :param function func: function to get its docstring.

    :param bool include_returns: specifies that `returns` and `rtype` parts
                                 of docstring must also be included in the result.
                                 defaults to True if not provided.

    :rtype: str
    """

    value = inspect.getdoc(func)
    if value is None:
        return None

    if include_returns is False:
        return_key = ':return'
        rtype_key = ':rtype'

        not_found_index = 5000000
        return_key_index = not_found_index
        rtype_key_index = not_found_index

        if return_key in value:
            return_key_index = value.index(return_key)

        if rtype_key in value:
            rtype_key_index = value.index(rtype_key)

        last_index = min(return_key_index, rtype_key_index)

        if last_index < not_found_index:
            value = value[0:last_index]

    return value


def get_fully_qualified_name(func):
    """
    gets the fully qualified name of given function.

    it returns `module_name.function_name`.
    for example: `pyrin.api.services.create_route`.

    :param function func: function to get its fully qualified name.
                          it must be a stand-alone function, not an
                          instance or class method.

    :rtype: str
    """

    return '{module}.{name}'.format(module=func.__module__, name=func.__name__)


def get_inputs(func, args, kwargs, container=dict, **options):
    """
    gets given function inputs as a dict from provided args and kwargs.

    it also returns the parent class or instance of function if available.

    :param function func: function to get its inputs.
    :param tuple args: a tuple of function positional inputs.
    :param dict kwargs: a dictionary of function keyword arguments.
    :param type container: a dict like object to be used for returning inputs.
                           defaults to dict if not provided.

    :returns: tuple[dict inputs, object | type parent]
    :rtype: tuple[dict, object | type]
    """

    if len(args) == 0 and len(kwargs) == 0:
        return container(), None

    kwargs_name = None
    signature = inspect.signature(func)
    for name, param in signature.parameters.items():
        if param.kind == Parameter.VAR_KEYWORD:
            kwargs_name = name
            break

    bounded_args = signature.bind_partial(*args, **kwargs)
    parent = bounded_args.arguments.pop('self', None)
    parent = bounded_args.arguments.pop('cls', parent)

    if kwargs_name is not None:
        keywords = bounded_args.arguments.pop(kwargs_name, None)
        if keywords is not None:
            bounded_args.arguments.update(keywords)

    return container(bounded_args.arguments), parent


def get_required_arguments(func):
    """
    gets all required arguments of given function.

    :param function func: function to get its required arguments names.

    :rtype: set[str]
    """

    signature = inspect.signature(func)
    result = []
    for name, item in signature.parameters.items():
        if item.kind in (Parameter.POSITIONAL_ONLY,
                         Parameter.POSITIONAL_OR_KEYWORD,
                         Parameter.KEYWORD_ONLY):

            if item.default is Parameter.empty:
                result.append(name)

    return set(result)


def apply(func, args, kwargs):
    """
    a helper method that applies args and kwargs to given function and gets the result.

    :param function func: function to be used.
    :param tuple args: function positional arguments.
    :param dict kwargs: function keyword arguments.

    :returns: function result
    """

    return func(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
database paging manager module.
"""

import pyrin.configuration.services as config_services

from pyrin.core.structs import Manager
from pyrin.database.paging import DatabasePagingPackage


class DatabasePagingManager(Manager):
    """
    database paging manager class.
    """

    # limit and offset keys that will be injected into
    # options of each view function if required.
    LIMIT_KEY = '__limit__'
    OFFSET_KEY = '__offset__'

    # paging parameters that application expects in query strings of requests.
    # these values could be customized in 'paging' section of database config store.
    PAGE_PARAM = 'page'
    PAGE_SIZE_PARAM = 'page_size'

    package_class = DatabasePagingPackage

    def __init__(self):
        """
        initializes an instance of DatabasePagingManager.
        """

        super().__init__()

        page_param = config_services.get('database', 'paging', 'page_param')
        if page_param not in (None, '') and not page_param.isspace():
            self._page_param = page_param
        else:
            self._page_param = self.PAGE_PARAM

        page_size_param = config_services.get('database', 'paging', 'page_size_param')
        if page_size_param not in (None, '') and not page_size_param.isspace():
            self._page_size_param = page_size_param
        else:
            self._page_size_param = self.PAGE_SIZE_PARAM

    def extract_paging_params(self, values):
        """
        extracts paging parameters from given dict and returns them as a new dict.

        the values will be removed from input dict.

        :param dict values: a dict to extract paging params from it.

        :returns: dict(int page: page number,
                       int page_size: page size)
        :rtype: dict
        """

        page = values.pop(self._page_param, None)
        page_size = values.pop(self._page_size_param, None)

        if isinstance(page, list):
            if len(page) > 0:
                page = page[0]
            else:
                page = None

        if isinstance(page_size, list):
            if len(page_size) > 0:
                page_size = page_size[0]
            else:
                page_size = None

        result = dict()
        result[self._page_param] = page
        result[self._page_size_param] = page_size
        return result

    def get_paging_params(self, **options):
        """
        gets paging parameters from given inputs.

        note that this method does not do any validation and just returns
        keys as they are, even if their value is None.

        :keyword int page: page number.
        :keyword int page_size: page size.

        :returns: tuple[int page, int page_size]
        :rtype: tuple[int, int]
        """

        return options.get(self._page_param), options.get(self._page_size_param)

    def generate_paging_params(self, page, page_size):
        """
        generates paging parameters from given inputs.

        :param int page: page number.
        :param int page_size: page size.

        :returns: dict[int page, int page_size]
        :rtype: dict[int, int]
        """

        params = dict()
        params[self._page_param] = page
        params[self._page_size_param] = page_size
        return params

    def inject_paging_keys(self, limit, offset, values):
        """
        injects paging keys into given dict.

        :param int limit: limit.
        :param int offset: offset.
        :param dict values: a dict to inject paging keys into it.
        """

        values[self.LIMIT_KEY] = limit
        values[self.OFFSET_KEY] = offset

    def get_paging_keys(self, **options):
        """
        gets paging keys from given inputs.

        note that this method does not do any validation and just returns
        keys as they are, even if their value is None.

        :keyword int __limit__: limit value.
        :keyword int __offset__: offset value.

        :returns: tuple[int limit, int offset]
        :rtype: tuple[int, int]
        """

        return options.get(self.LIMIT_KEY), options.get(self.OFFSET_KEY)

    def disable_paging_keys(self, values):
        """
        disables paging keys in given dict.

        :param dict values: a dict to disable paging keys in it.
        """

        self.inject_paging_keys(limit=None, offset=None, values=values)



# -*- coding: utf-8 -*-
"""
celery audit manager module.
"""

import traceback

from pyrin.core.structs import Manager
from pyrin.audit.enumerations import InspectionStatusEnum
from pyrin.task_queues.celery.audit import CeleryAuditPackage
from pyrin.task_queues.celery.audit.tasks import audit_task


class CeleryAuditManager(Manager):
    """
    celery audit manager class.
    """

    package_class = CeleryAuditPackage

    def __init__(self):
        """
        initializes an instance of CeleryAuditManager.
        """

        super().__init__()

    def inspect(self, **options):
        """
        inspects the status of celery.

        it returns a tuple of two values. first value is a dict containing the inspection
        data. and the second value is a bool value indicating that inspection has been
        succeeded or failed.

        :keyword bool traceback: specifies that on failure report, it must include
                                 the traceback of errors.
                                 defaults to True if not provided.

        :keyword bool raise_error: specifies that it must raise error
                                   if any of registered audits failed
                                   instead of returning a failure response.
                                   defaults to False if not provided.

        :rtype: tuple[dict, bool]
        """

        raise_error = options.get('raise_error', False)
        include_traceback = options.get('traceback', True)
        data = {}
        succeeded = True
        try:
            audit_task.delay(suppress=False)
            data.update(status=InspectionStatusEnum.OK)
        except Exception as error:
            if raise_error is True:
                raise

            succeeded = False
            data.update(status=InspectionStatusEnum.FAILED,
                        error=str(error))
            if include_traceback is not False:
                data.update(traceback=traceback.format_exc())

        return data, succeeded

    def perform_job(self):
        """
        performs a dummy job.

        this method is implemented to be called by `audit_task` to check celery.
        """
        pass
